
/* U7RUN.CPP
  -----------

  Program to run the patched U7.EXE

  Copyright (C) 2000 by Josef Drexler <jdrexler@julian.uwo.ca>

  Distributed under the terms of the GNU General Public License.  For more
  information please read the file COPYING which is included in the
  distribution.


  Compiling:
  Set SHOWSTATS to 1 if you want to see memory access stats every second
*/


#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <process.h>
#include <time.h>

#include "u7patch.h"


#define INTPAR

#ifdef __cplusplus
#undef INTPAR
#define INTPAR ...
extern "C" {
#endif

#define SHOWSTATS 1

#define ACCESS ACCESS_FILE

#define USEBUFFER 1	// set to 0 for unbuffered accesses (might be faster...)

#define BUFFERSIZE 256	// size of each buffer
#define NBUFFER 1	// number of buffers. only one right now
#define BACKBUFFER (BUFFERSIZE/16)
			// bytes to accomodate decreasing-address accesses

#define COUNTDATA 1	// size or 1, to count bytes or accesses


#if ACCESS==ACCESS_XMS

#include "xms.h"

int xmshandle = 0;

#endif


void interrupt threebyteop(INTPAR), fourbyteop(INTPAR), stringop(INTPAR),
	movimm(INTPAR), xlatint(INTPAR), special(INTPAR);
void segfault(int code),savestack();

int voodoofile;
char* logpath;

typedef long unsigned int U32;

U32 stats[9];
U32 intcalls[INT_LAST - INT_FIRST + 2];

int intsbad = 0;
void far *oldints[INT_LAST - INT_FIRST + 1];

char* buffer;
#if USEBUFFER
U32 bufstart = 0;
int bufreadfrom = 0;
int bufwritten = 0;
#endif

time_t lasttime, firsttime;

void showstats(FILE *file)
{
  fprintf(file,"\n"
	  "U7RUN Statistics\n"
	  "----------------\n\n"
	  "Voodoo access:   Above 1MB  Below 1MB  Video Mem\n"
	  "         read: %11ld%11ld%11ld\n"
	  "        write: %11ld%11ld%11ld\n\n"
#if USEBUFFER
	  "Read page       %11ld (with reads: %ld)\n"
	  "Write dirty page%11ld\n\n"
#endif
	  "Interrupt Calls\n"
	  "A0-A3 %11ld %11ld %11ld %11ld\n"
	  "A4-A7 %11ld %11ld %11ld %11ld\n"
	  "A8-AB %11ld %11ld %11ld %11ld\n"
	  "AC-AF %11ld %11ld %11ld %11ld\n"
	  "B0-B3 %11ld %11ld %11ld %11ld\n"
	  "B4-B7 %11ld %11ld %11ld %11ld\n"
	  "B8-BB %11ld %11ld %11ld %11ld\n"
	  "BC-BF %11ld %11ld %11ld %11ld\n"
	  "C0    %11ld\n"
	  "Total %11ld after %d seconds\n",
	  stats[0], stats[2], stats[6], stats[1], stats[3], stats[7],
#if USEBUFFER
	  stats[4], stats[5], stats[8],
#endif
	  intcalls[0], intcalls[1], intcalls[2], intcalls[3],
	  intcalls[4], intcalls[5], intcalls[6], intcalls[7],
	  intcalls[8], intcalls[9], intcalls[10], intcalls[11],
	  intcalls[12], intcalls[13], intcalls[14], intcalls[15],
	  intcalls[16], intcalls[17], intcalls[18], intcalls[19],
	  intcalls[20], intcalls[21], intcalls[22], intcalls[23],
	  intcalls[24], intcalls[25], intcalls[26], intcalls[27],
	  intcalls[28], intcalls[29], intcalls[30], intcalls[31],
	  intcalls[32], intcalls[33],
	  lasttime - firsttime);
  fflush(file);
}

#if USEBUFFER
void tobuffer(U32 ofs)
{
  if (bufwritten) {	// write back dirty buffer
	stats[5]++;
#if ACCESS==ACCESS_FILE
	lseek(voodoofile, bufstart, SEEK_SET);
	if (write(voodoofile, buffer, BUFFERSIZE * NBUFFER) != NBUFFER*BUFFERSIZE) {
		perror("Writing");
		segfault(20);
	}
#elif ACCESS==ACCESS_XMS
	if (!xmsaccess(xmshandle, buffer, bufstart, BUFFERSIZE * NBUFFER, 1)) {
		segfault(6);
	}
#endif
	bufwritten = 0;
  }
  stats[4]++;
  if (bufreadfrom) {
	stats[8]++;
  }
  bufreadfrom = 0;
  bufstart = (ofs & ~3L);
  if (bufstart >= BACKBUFFER)
	bufstart -= BACKBUFFER;

#if ACCESS==ACCESS_FILE
  lseek(voodoofile, bufstart, SEEK_SET);
  if (read(voodoofile, buffer, NBUFFER*BUFFERSIZE) != NBUFFER*BUFFERSIZE) {
	perror("Reading");
	segfault(9);
  }
#elif ACCESS==ACCESS_XMS
  if (!xmsaccess(xmshandle, buffer, bufstart, BUFFERSIZE * NBUFFER, 0)) {
	segfault(7);
  }
#endif


  if (time(NULL) != lasttime) {
	lasttime = time(NULL);
#if SHOWSTATS
	asm {
		mov ax,3
		int 10h
	}
	showstats(stdout);
#endif
  }
}

U32 bufpos(U32 ofs)
{
  U32 newofs;
  if ( (ofs < bufstart) || (ofs + 4 > bufstart + BUFFERSIZE) )
	tobuffer(ofs);
  newofs = ofs - bufstart;
  if (newofs + 4 > bufstart + BUFFERSIZE) {
	segfault(8);
  }
  return newofs;
}
#endif

U32 getvoodoomem(U32 ofs, int size)
{
  U32 value;

  if (ofs < MEM1MB) {		// it's real mode memory!
	stats[0] += COUNTDATA;
	if ( (ofs >= 0xa0000) && (ofs < 0xb0000) )
		stats[6] += COUNTDATA;
	asm {
		mov eax,ofs	// access real memory to be consistent
		mov di,ax	// with non-voodoo access to the same
		and di,15	// memory
		shr eax,4
		push es
		mov es,ax
		mov eax,es:[di]
		pop es
		mov value,eax
	}
  } else {			// read from the emulated memory
	ofs -= MEM1MB;
	stats[1] += COUNTDATA;
	if (ofs > VOODOOSIZE) {
		segfault(2);
	}
#if USEBUFFER
	asm {
		push dword ptr ofs
		call bufpos
		add sp,4
#if defined(__TINY__) || defined(__SMALL__) || defined(__MEDIUM__)
		mov di,buffer
		add di,ax
		mov eax,[di]
#else
		push es
		les di,dword ptr buffer
		add di,ax
		mov eax,es:[di]
		pop es
#endif
		mov value,eax
		mov bufreadfrom,1;
	}
#elif ACCESS==ACCESS_FILE
	lseek(voodoofile, ofs, SEEK_SET);
	read(voodoofile, &value, 4);
#else
#	error Invalid combination.
#endif
  }

  return value;
}

void setvoodoomem(U32 ofs, U32 value, int size)
{
  if ( (size != 1) && (size != 2) && (size != 4) ) {
	segfault(5);
  }
  if (ofs < MEM1MB) {		// it's real mode memory!
	stats[2] += COUNTDATA;
	if ( (ofs >= 0xa0000) && (ofs < 0xb0000) )
		stats[7] += COUNTDATA;
	// this was a result of a bug... don't need it really...
#if 1
	if (ofs < 0x400) {
		// it's system memory...  not good...
		segfault(4);;
	} else
#endif
	asm {
		push si
		mov eax,ofs	// access the real memory to be consistent
		mov di,ax
		and di,15
		shr eax,4
		push es
		push ds
		mov es,ax
		mov ax,ss
		mov ds,ax
		lea si,value
		mov cx,size
		rep movsb	// need movsb because then we can set the size
		pop ds
		pop es
		pop si
	}
  } else {
		// write to the emulated memory
	ofs -= MEM1MB;
	stats[3] += COUNTDATA;
	if (ofs > VOODOOSIZE) {
		segfault(3);
	}
#if USEBUFFER
	asm {
		push es
		push si
		push dword ptr ofs
		call bufpos
		add sp,4
#if defined(__TINY__) || defined(__SMALL__) || defined(__MEDIUM__)
		mov bx,ds
		mov es,bx
		mov di,buffer
#else
		les di,dword ptr buffer
#endif
		add di,ax
		cld
		push ds
		mov ax,ss
		mov ds,ax
		lea si,value
		mov cx,size
		rep movsb
		pop ds
		pop si
		pop es
		mov bufwritten,1
	}
#elif ACCESS==ACCESS_FILE
	lseek(voodoofile, ofs, SEEK_SET);
	write(voodoofile, &value, size);
	if (time(NULL) != lasttime) {
		lasttime = time(NULL);
#		if SHOWSTATS
			asm {
				mov ax,3
				int 10h
			}
			showstats(stdout);
#		endif
  }
#else
#	error Invalid combination.
#endif
/*
	fseek(voodoofile, ofs, SEEK_SET);
	fwrite(&value, size, 1, voodoofile);
	asm {
;		push es
;		mov ax,40h
;		mov es,ax
;		and byte ptr es:[17h],not 64
;		xor byte ptr es:[17h],16
;		pop es
	}
*/
  }
}

#ifdef __cplusplus
}
#endif

void vsetvect(int interruptno, void far *handler)
{
  typedef void interrupt (*interrupthandler)(INTPAR);
  interrupthandler ihandler;

  // stupid borland c doesn't allow the necessary type casting...
  asm {
	mov eax,dword ptr handler
	mov dword ptr ihandler,eax
  }
  setvect(interruptno, ihandler);
}

void restoreints()
{
  int intnum;

  if (!intsbad)
	return;

  for (intnum = INT_FIRST; intnum <= INT_LAST; intnum++)
	vsetvect(intnum, oldints[intnum - INT_FIRST]);

  intsbad = 0;
}

void saveints()
{
  int intnum;

  for (intnum = INT_FIRST; intnum <= INT_LAST; intnum++)
	oldints[intnum - INT_FIRST] = getvect(intnum);
}

void newints()
{
  int intnum;

  intsbad = 1;
  atexit(restoreints);

  for (intnum = INT_MOV_3; intnum < INT_MOV_4; intnum++)
	setvect(intnum, threebyteop);
  for (intnum = INT_MOV_4; intnum < INT_SOP; intnum++)
	setvect(intnum, &fourbyteop);
  for (intnum = INT_SOP; intnum < INT_MOVIMM; intnum++)
	setvect(intnum, &stringop);
  setvect(INT_MOVIMM, &movimm);
  setvect(INT_XLAT, &xlatint);
  setvect(INT_SPCL, &special);
}

#if ACCESS==ACCESS_XMS
void freexms()
{
  if (xmshandle)
	xmsfree(xmshandle);

  xmshandle = 0;
}
#endif

#if ACCESS==ACCESS_FILE
int makefile()
{
  char* temppath = getenv("TEMP");
  if (!temppath)
	temppath = getenv("TMP");
  if (!temppath)
	temppath = "C:\\";
  char* filename = "%s\\U7RUN%02x.TMP";
  int file = -1;

  if (logpath[strlen(logpath)-1] == '\\') {
	filename++;	// temp path already has a trailing \, so skip
  }			// the one given in filename

  logpath = (char*) malloc(strlen(temppath) + strlen(filename) + 1);

  for (int lognum=0; (lognum < 0x100) && (file == -1); lognum++) {
	sprintf(logpath, filename, temppath, lognum);

	_fmode = O_BINARY;
	file = creat(logpath, S_IREAD | S_IWRITE);
  }

  if (file == -1) {
	fprintf(stderr, "Cannot open %s: %s\n", logpath, strerror(errno));
	exit(1);
  }

  printf("Creating %s...\n", logpath);

	// make a file of 2MB, filled with zeroes
  for (unsigned int i=0; i<(VOODOOSIZE / NBUFFER / BUFFERSIZE); i++)
	if (write(file, buffer, NBUFFER * BUFFERSIZE) != NBUFFER * BUFFERSIZE) {
		perror("Writing to the memory file");
		exit(1);
	}

  lseek(file, 0, SEEK_SET);
  if (read(file, buffer, BUFFERSIZE) != BUFFERSIZE) {
	perror("Reading from the memory file");
	exit(1);
  }

  return file;
}
#endif

void init()
{
  lasttime = firsttime = time(NULL);
  memset(stats, 0, sizeof(stats));
  memset(intcalls, 0, sizeof(intcalls));
  buffer = (char*) calloc(NBUFFER*BUFFERSIZE, 1);

#if ACCESS==ACCESS_FILE
  voodoofile = makefile();
#elif ACCESS==ACCESS_XMS
  if (!xmsinit()) {
	printf("Cannot initialize XMS memory (code %x)\n", xmserror);
	exit(1);
  }
  int xmsavail = xmsmemfree();
  if (xmsavail < VOODOOSIZE / 1024) {
	printf("Not enough XMS memory available (code %x), Have %d KB, need %d KB.\n",
		xmserror, xmsavail, VOODOOSIZE / 1024);
	exit(1);
  }
  xmshandle = xmsalloc(VOODOOSIZE / 1024);
  if (!xmshandle) {
	printf("Could not allocate XMS memory (code %x).\n", xmserror);
	exit(1);
  }
  atexit(freexms);
#endif
}

void cleanup()
{
#if ACCESS==ACCESS_FILE
  close(voodoofile);
#elif ACCESS==ACCESS_XMS
  freexms();
#endif
  if (buffer)
	free(buffer);
}

#if USEBUFFER
void testmemory()
{
  U32 ofs;
  int num, value, bofs;

  printf("Testing buffered memory...\n");
  // do 256 random write accesses
  srand(2344);
  for (num=0; num<256; num++) {
	printf("\rWriting: %3d/256", num);
	value = random(255);
	ofs = ((1L * random(VOODOOSIZE/65536L)) << 16L) + random(32767);
	bofs = bufpos(ofs);
	if ( (bofs < 0) || (bofs >= NBUFFER * BUFFERSIZE) ) {
		printf("bufpos() isn't working properly!\n");
		exit(1);
	}
	buffer[bofs] = value;
	bufwritten = 1;
  }
  printf("\rWriting: done   \n");

  srand(2344);	// initialize randomizer with same value as before
  for (num=0; num<256; num++) {
	printf("\rReading: %3d/256", num);
	value = random(255);
	ofs = ((1L * random(VOODOOSIZE/65536L)) << 16L) + random(32767);
	bofs = bufpos(ofs);
	if (buffer[bofs] != value) {
		printf("\nMemory is leaking bits!\n");
		exit(1);
	}
	bufreadfrom = 1;
  }
  printf("\rReading: done   \n");
}
#endif

int main()
{
  init();

#if USEBUFFER
  testmemory();
#endif

	// run U7.  not using system() because we can't afford the memory
	// command.com would steal
  saveints();
  newints();
//  segfault(0x3456);	// for testing
  savestack();
  spawnl(P_WAIT, "U7.EXE", "U7", "p", NULL);
//  spawnl(P_WAIT, getenv("COMSPEC"), "command", "/c", "mem","/d","/p", NULL);

  restoreints();

  cleanup;

  lasttime = time(NULL);
  printf("\nPress any key to show statistics.\n");
  asm {
	mov ah,0
	int 16h
  }

  showstats(stdout);

  printf("\nPress any key to exit.\n");
  asm {
	mov ah,0
	int 16h
  }

  return 0;
}
